//******************************************************************************
//*
//*     FULLNAME:  Single-Chip Microcontroller Real-Time Operating System
//*
//*     NICKNAME:  scmRTOS
//*
//*     PROCESSOR: ARM Cortex-M0, Nordic Semiconductor nRF51822
//*
//*     TOOLKIT:   EWARM (IAR Systems)
//*               
//*     PURPOSE:   Target Dependent Low-Level Stuff
//*               
//*     Version: 5.1.0
//*
//*     $Revision: 1 $
//*     $Date:: 2016-12-15 #$
//*
//*     Copyright (c) 2003-2012, Harry E. Zhurov
//*
//*     Permission is hereby granted, free of charge, to any person 
//*     obtaining  a copy of this software and associated documentation 
//*     files (the "Software"), to deal in the Software without restriction, 
//*     including without limitation the rights to use, copy, modify, merge, 
//*     publish, distribute, sublicense, and/or sell copies of the Software, 
//*     and to permit persons to whom the Software is furnished to do so, 
//*     subject to the following conditions:
//*
//*     The above copyright notice and this permission notice shall be included 
//*     in all copies or substantial portions of the Software.
//*
//*     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//*     EXPRESS  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
//*     MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
//*     IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
//*     CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
//*     TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
//*     THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//*
//*     =================================================================
//*     See http://scmrtos.sourceforge.net for documentation, latest
//*     information, license and contact details.
//*     =================================================================
//*
//******************************************************************************
//* Cortex-M0 nRF51x IAR port, Akimov Vladimir, 'sevstels', Copyright (c) 2016

#include "scmRTOS_CONFIG.h"
#include "scmRTOS_TARGET_CFG.h"

//------------------------------------------------------------------------------
//  PUBLIC FUNCTIONS
//
    EXTERN  os_context_switch_hook
    EXTERN  os_init_custom_system_timer
    PUBLIC  os_start
    PUBLIC  PendSV_Handler

//------------------------------------------------------------------------------
//  EQUATES
NVIC_INT_CTRL        EQU  0xE000ED04  // ICSR Interrupt control state register.
NVIC_PENDSVSET       EQU  0x10000000  // Value to trigger PendSV exception.

SHP_ADDRESS          EQU  0xE000ED20
NVIC_PENDSV_PRI_W    EQU  0x00FF0000 

//------------------------------------------------------------------------------
//    CODE GENERATION DIRECTIVES
//
    RSEG CODE:CODE(2)
    THUMB

//------------------------------------------------------------------------------
//    HANDLE PendSV EXCEPTION
//    void PendSV_Handler(void)
//
// Note(s) : 
// 1) PendSV is used to cause a context switch.  This is a recommended method 
//    for performing context switches with Cortex-M3.  This is because the 
//    Cortex-M3 auto-saves half of the processor context on any exception, 
//    and restores same on return from exception.  So only saving of R4-R11 
//    is required and fixing up the stack pointers.  Using the PendSV exception
//    this way means that context saving and restoring is identical whether it 
//    is initiated from a thread or occurs due to an interrupt or exception.
//
// 2) Pseudo-code is:
//    a) Get the process SP
//    b) Save remaining regs r4-r11 on process stack;
//    c) Call os_context_switch_hook for save current task SP and get new task SP;
//    d) Restore R4-R11 from new process stack;
//    e) Perform exception return which will restore remaining context.
//
//    3) On entry into PendSV handler:
//    a) The following have been saved on the process stack (by processor):
//       xPSR, PC, LR, R12, R0-R3
//    b) Processor mode is switched to Handler mode (from Thread mode)
//    c) Stack is Main stack (switched from Process stack)
//
//    4) Since PendSV is set to lowest priority in the system 
//      (by os_start() below), we know that it will only be run when no other 
//      exception or interrupt is active, and therefore safe to assume that 
//      context being switched out was using the process stack (PSP).
//

/*
 * PendSV is used to perform a context switch. This is a recommended method for Cortex-M.
 * This is because the Cortex-M automatically saves half of the processor context
 * on any exception, and restores same on return from exception.  So only saving of R4-R11
 * and fixing up the stack pointers is required.  Using the PendSV exception this way means
 * that context saving and restoring is identical whether it is initiated from a thread
 * or occurs due to an interrupt or exception.
 *
 * Since PendSV interrupt has the lowest priority in the system (set by os_start() below),
 * we can be sure that it will run only when no other exception or interrupt is active, and
 * therefore safe to assume that context being switched out was using the process stack (PSP).
 */
 
// For Cortex-M0(M0+)/Cortex-M1
PendSV_Handler
  
  // Prevent interruption during context switch
  CPSID   I  
  
  // Load process stack pointer to R0
  MRS   R0, PSP
  
  // Adjust R0 to point to top of saved context in stack
  SUBS  R0, R0, #32
  
  // Preserve R0 (needed for os_context_switch_hook() call)
  MOV   R1, R0
  
  // Save low portion of remaining registers (r4-7) on process stack
  STMIA R1!, {R4-R7}
  
  // Move high portion of remaining registers (r8-11) to low registers
  MOV   R4, R8  
  MOV   R5, R9 
  MOV   R6, R10 
  MOV   R7, R11
  
  // Save high portion of remaining registers (r8-11) on process stack
  STMIA R1!, {R4-R7}

  // At this point, entire context of process has been saved
  // we must save LR (exc_return value) until exception return
  PUSH {LR}
  
  // Call os_context_switch_hook();
  LDR   R1, =os_context_switch_hook   
  BLX   R1

  // R0 is new process SP;
  // Adjust R0 to point to high registers (r8-11)        
  ADDS  R0, R0, #16 
  
  // Restore r8-11 from new process stack
  LDMIA R0!, {R4-R7}
  
  // Move restored values to high registers (r8-11)
  MOV   R8, R4   
  MOV   R9, R5 
  MOV   R10, R6
  MOV   R11, R7
  
  // R0 at this point is new process SP
  MSR   PSP, R0
  
  // Adjust R0 to point to low registers
  SUBS  R0, R0, #32
  
  // Restore r4-7
  LDMIA R0!, {R4-R7} 
  CPSIE I 
  
  // Return to saved exc_return. 
  // Exception return will restore remaining context
  POP  {PC}


//------------------------------------------------------------------------------
// START MULTITASKING
// void os_start(stack_item_t* sp)
//
// Note(s) : 1) os_start() MUST:
// a) Setup PendSV and SysTick exception priority to lowest;
// b) Setup SysTick (reload value);
// c) Enable interrupts (tasks will run with interrupts enabled).
// d) Jump to exec() function of the highest priority process.

// For Cortex-M0 nRF51x
os_start

  // Set the PendSV exception priority (lowest)
  LDR   R1, =SHP_ADDRESS
  LDR   R2, =NVIC_PENDSV_PRI_W
  STR   R2, [R1]
    
  // Load process entry point into R4, by address from R0 with offset 56 bytes 
  LDR   R4, [R0, #(4 * 14)] 
   
  // Emulate context restore
  ADDS  R0, R0, #(4 * 16)       

  // Store process SP to PSP
  MSR   PSP, R0 
  
  // Switch thread mode stack to PSP
  MOVS  R0, #2                  
  
  // 
  MSR   CONTROL, R0
  
  // Insert a barrier
  ISB
 
  //For Nordic nRF51x chip use Timer1 like SysTic
  #if scmRTOS_USE_CUSTOM_SYSTIMER != 0

  // Init and run system timer
  LDR   R1, =os_init_custom_system_timer      
  
  //Jump to init function
  BLX   R1

  #endif

  // Enable interrupts at processor level
  CPSIE I                       
  
  // Jump to process exec() function
  BX    R4 

  END
