//******************************************************************************
//*
//*     FULLNAME:  Single-Chip Microcontroller Real-Time Operating System
//*
//*     NICKNAME:  scmRTOS
//*
//*     PROCESSOR: ARM Cortex-M0, Nordic Semiconductor nRF51822
//*
//*     TOOLKIT:   EWARM (IAR Systems)
//*
//*     PURPOSE:   Target Dependent Stuff Source
//*
//*     Version: 5.00
//*
//*     $Revision: 1 $
//*     $Date:: 2016-12-15 #$
//*
//*     Copyright (c) 2003-2012, Harry E. Zhurov
//*
//*     Permission is hereby granted, free of charge, to any person
//*     obtaining  a copy of this software and associated documentation
//*     files (the "Software"), to deal in the Software without restriction,
//*     including without limitation the rights to use, copy, modify, merge,
//*     publish, distribute, sublicense, and/or sell copies of the Software,
//*     and to permit persons to whom the Software is furnished to do so,
//*     subject to the following conditions:
//*
//*     The above copyright notice and this permission notice shall be included
//*     in all copies or substantial portions of the Software.
//*
//*     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//*     EXPRESS  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//*     MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//*     IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//*     CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//*     TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
//*     THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//*
//*     =================================================================
//*     See http://scmrtos.sourceforge.net for documentation, latest
//*     information, license and contact details.
//*     =================================================================
//*
//******************************************************************************
//* Cortex-M3/M4(F) GCC port by Anton B. Gusev aka AHTOXA, Copyright (c) 2012
//* Cortex-M0 GCC port by Sergey A. Borshch, Copyright (c) 2011

#include <scmRTOS.h>

using namespace OS;

namespace{
// Cortex-M0/M0+/M3/M4 context size = 16*4 = 64 bytes
enum { CONTEXT_SIZE = 16 * sizeof(stack_item_t)};
}

void TBaseProcess::init_stack_frame( stack_item_t * Stack
                                   , void (*exec)()
                                #if scmRTOS_DEBUG_ENABLE == 1
                                   , stack_item_t * StackBegin
                                #endif
                                   )
{
 //ARM Architecture Procedure Call Standard [AAPCS] requires 8-byte stack alignment.
 //This means that we must get top of stack aligned _after_ context "pushing", at
 //interrupt entry.

 uintptr_t sptr = (((uintptr_t)Stack-CONTEXT_SIZE)&0xFFFFFFF8UL) + CONTEXT_SIZE;
 StackPointer = (stack_item_t*)sptr;

 // Write to last 4 bytes
 *(--StackPointer)  = 0x01000000UL; // xPSR
    
 // Entry Point for () 
 *(--StackPointer) = reinterpret_cast<uint32_t>(exec); 

 // Core without FPU, emulate "push LR,R12,R3,R2,R1,R0,R11-R4"
 StackPointer -= 14;

#if scmRTOS_DEBUG_ENABLE == 1
    for (stack_item_t* pDst = StackBegin; pDst < StackPointer; pDst++)
        {*pDst = STACK_DEFAULT_PATTERN;}
#endif // scmRTOS_DEBUG_ENABLE
}

// weak alias to support old name of PendSV_Handler.
#pragma weak PendSVC_ISR = PendSV_Handler

//Some Cortex-M0 registers and constants.
namespace
{

template<uint32_t addr, typename type = uint32_t>
struct ioregister_t
{
  type operator=(type value) { *(volatile type*)addr = value; return value; }
  void operator|=(type value) { *(volatile type*)addr |= value; }
  void operator&=(type value) { *(volatile type*)addr &= value; }
  operator type() { return *(volatile type*)addr; }
};

template<uint32_t addr, class T>
struct iostruct_t
{
  volatile T* operator->() { return (volatile T*)addr; }
};

// SysTick stuff
struct systick_t
{
  uint32_t CTRL;
  uint32_t LOAD;
  uint32_t VAL;
  uint32_t const CALIB;
};

enum
{
  NVIC_ST_CTRL_CLK_SRC = 0x00000004,       // Clock Source.
  NVIC_ST_CTRL_INTEN   = 0x00000002,       // Interrupt enable.
  NVIC_ST_CTRL_ENABLE  = 0x00000001        // Counter mode.
};

////static iostruct_t<0xE000E010UL, systick_t> SysTick;

//Cortex-M0  Floating point context control register stuff
enum
{
  ASPEN = (0x1UL << 31),  // always save enable
  LSPEN = (0x1UL << 30)   // lazy save enable
};

}

/*
 * System timer stuff.
 * This is the default (weak) system timer ISR handler.
 * The user can redefine this handler if needed.
 */
//#pragma weak SysTick_Handler = Default_SystemTimer_ISR
//#pragma weak SystemTimer_ISR = Default_SystemTimer_ISR
/*
#pragma weak TIMER1_IRQHandler = Default_SystemTimer_ISR
namespace OS
{
extern "C" OS_INTERRUPT void Default_SystemTimer_ISR()
{
    scmRTOS_ISRW_TYPE ISR;

#if scmRTOS_SYSTIMER_HOOK_ENABLE == 1
    system_timer_user_hook();
#endif

    Kernel.system_timer();

#if scmRTOS_SYSTIMER_NEST_INTS_ENABLE == 0
    DISABLE_NESTED_INTERRUPTS();
#endif
}
}*/

#if scmRTOS_PRIORITY_ORDER == 0
namespace OS
{
 extern TPriority const PriorityTable[] =
 {
   #if scmRTOS_PROCESS_COUNT == 1
       (TPriority)0xFF,
       pr0,
       prIDLE, pr0
   #elif scmRTOS_PROCESS_COUNT == 2
       (TPriority)0xFF,
       pr0,
       pr1, pr0,
       prIDLE, pr0, pr1, pr0
   #elif scmRTOS_PROCESS_COUNT == 3
       (TPriority)0xFF,
       pr0,
       pr1, pr0,
       pr2, pr0, pr1, pr0,
       prIDLE, pr0, pr1, pr0, pr2, pr0, pr1, pr0
   #elif scmRTOS_PROCESS_COUNT == 4
       (TPriority)0xFF,
       pr0,
       pr1, pr0,
       pr2, pr0, pr1, pr0,
       pr3, pr0, pr1, pr0, pr2, pr0, pr1, pr0,
       prIDLE, pr0, pr1, pr0, pr2, pr0, pr1, pr0, pr3, pr0, pr1, pr0, pr2, pr0, pr1, pr0
    #elif scmRTOS_PROCESS_COUNT == 5
       (TPriority)0xFF,
       pr0,
       pr1, pr0,
       pr2, pr0, pr1, pr0,
       pr3, pr0, pr1, pr0, pr2, pr0, pr1, pr0,
       pr4, pr0, pr1, pr0, pr2, pr0, pr1, pr0, pr3, pr0, pr1, pr0, pr2, pr0, pr1, pr0,
       prIDLE, pr0, pr1, pr0, pr2, pr0, pr1, pr0, pr3, pr0, pr1, pr0, pr2, pr0, pr1, pr0, pr4, pr0, pr1, pr0, pr2, pr0, pr1, pr0, pr3, pr0, pr1, pr0, pr2, pr0, pr1, pr0
     #else // scmRTOS_PROCESS_COUNT > 5
       (TPriority)32,   (TPriority)0,   (TPriority)1,   (TPriority)12,
       (TPriority)2,    (TPriority)6,   (TPriority)0xFF,(TPriority)13,
       (TPriority)3,    (TPriority)0xFF,(TPriority)7,   (TPriority)0xFF
       (TPriority)0xFF, (TPriority)0xFF,(TPriority)0xFF,(TPriority)14,
       (TPriority)10,   (TPriority)4,   (TPriority)0xFF,(TPriority)0xFF,
       (TPriority)8,    (TPriority)0xFF,(TPriority)0xFF,(TPriority)25,
       (TPriority)0xFF, (TPriority)0xFF,(TPriority)0xFF,(TPriority)0xFF,
       (TPriority)0xFF, (TPriority)21,  (TPriority)27,  (TPriority)15,
       (TPriority)31,   (TPriority)11,  (TPriority)5,   (TPriority)0xFF,
       (TPriority)0xFF, (TPriority)0xFF,(TPriority)0xFF,(TPriority)0xFF,
       (TPriority)9,    (TPriority)0xFF,(TPriority)0xFF,(TPriority)24,
       (TPriority)0xFF, (TPriority)0xFF,(TPriority)20,  (TPriority)26,
       (TPriority)30,   (TPriority)0xFF,(TPriority)0xFF,(TPriority)0xFF,
       (TPriority)0xFF, (TPriority)23,  (TPriority)0xFF,(TPriority)19,
       (TPriority)29,   (TPriority)0xFF,(TPriority)22,  (TPriority)18,
       (TPriority)28,   (TPriority)17,  (TPriority)16,  (TPriority)0xFF
     #endif // scmRTOS_PROCESS_COUNT
    };
} //namespace
#endif
